module Horn (splitIf, trueSymbol, falseSymbol, trueTerm, falseTerm) where

import Data.List
import Text.Read
import Term
import Equation
import ES
import TPTP

-- Checking if a Horn cluase admits a trivial model.

type HornClause a = ([a], Maybe a)
-- E.g.,
--   ([1,2,3], Just 4)  stands for "(1 && 2 && 3) -> 4".
--   ([1,2,3], Nothing) stands for "(1 && 2 && 3) -> False".

-- Conditional equations (used as intermediate data)
type CEquation = ([Equation], Equation)

type CES = [CEquation]

-- Satisfiability checker of propositional Horn formulas

deleteAll :: Eq a => a -> [a] -> [a]
deleteAll x ys = [ y | y <- ys, y /= x ]

satisfiable :: Eq a => [HornClause a] -> Bool
satisfiable cs = satisfiable1 cs cs

satisfiable1 cs1 []                   = True
satisfiable1 cs1 (([], Nothing) : cs) = False
satisfiable1 cs1 (([], y@(Just a)) : cs) = satisfiable cs2 
  where cs2 = [ (deleteAll a xs, z) | (xs, z) <- cs1, y /= z ]
satisfiable1 cs1 (_ : cs) = satisfiable1 cs1 cs

encodeTPTP :: TPTP -> Maybe [HornClause Atom]
encodeTPTP tptp = sequenceA [ encodeClause clause | CNF _ _ clause <- tptp ]

encodeClause []                               = Just ([], Nothing)
encodeClause ((True, z) : clause)
  | Just (xs, Nothing) <- encodeClause clause = Just (xs, Just z)
encodeClause ((False, z) : clause)
  | Just (xs, ys) <- encodeClause clause      = Just (z : xs, ys)
encodeClause _ = Nothing 


rootOfPredicate (Predicate (F f _)) = Just f
rootOfPredicate (Equation _ _)      = Nothing

propositionalEncoding hornClauses =
  [ ([ f | Predicate (F f _) <- ts ], Nothing)
  | (ts, Nothing) <- hornClauses ] ++
  [ ([ f | Predicate (F f _) <- ts ], Just g)
  | (ts, Just (Predicate (F g _))) <- hornClauses ]

hasTrivialModel :: [HornClause Atom] -> Bool
hasTrivialModel hornClauses = satisfiable (propositionalEncoding hornClauses)

-- Trasformation from Horn clauses to word problems
-- (Claessen and Smallbone, IJCAR 2018)

trueSymbol  = "true__"
falseSymbol = "false__"
trueTerm    = F trueSymbol []
falseTerm   = F falseSymbol []

liftAtom :: Atom -> (Term, Term)
liftAtom (Predicate t)  = (t, F trueSymbol [])
liftAtom (Equation s t) = (s, t)

liftHornClause :: HornClause Atom -> CEquation
liftHornClause (ps, Nothing) =
  ([ liftAtom p | p <- ps], (F falseSymbol [], F trueSymbol []))
liftHornClause (ps, Just e)  =
  ([ liftAtom p | p <- ps], liftAtom e)

liftHornClauses :: [HornClause Atom] -> CES
liftHornClauses hornClauses = [ liftHornClause c | c <- hornClauses ]

fresh :: Int -> String
fresh n = "f" ++ show n

functionsInCES ces = functionsInES [ e' | (es, e) <- ces, e' <- e : es ]

fSuffix :: String -> Maybe Int
fSuffix ('f' : s) = readMaybe s
fSuffix _         = Nothing

fIndex ces =
  foldl max 0 [ n | f <- functionsInCES ces, Just n <- [fSuffix f] ] + 1

splitIfCES' :: Int -> CES -> ES
splitIfCES' n [] = []
splitIfCES' n (([], e) : hornClauses) = e : splitIfCES' n hornClauses
splitIfCES' n (((s0, t0) : es, (u, v)) : hornClauses) =
  (F f (s : ws), u) : splitIfCES' (n + 1) ((es, (F f (t : ws), v)) : hornClauses)
  where
    f = fresh n
    ws = [ V x | x <- nub (variables s ++ variables u ++ variables v) ]
    (s, t) = if length (variables s0) < length (variables t0) then (s0, t0) else (t0, s0)

splitIfCES :: CES -> ES
splitIfCES hornClauses = splitIfCES' (fIndex hornClauses) hornClauses

-- ES: axioms
-- Equation: goal
splitIf :: TPTP -> Maybe (ES, Equation)
splitIf tptp
  | Just hornClauses <- encodeTPTP tptp =
      if hasTrivialModel hornClauses then
        Just ([], (trueTerm, trueTerm))
      else
        Just (splitIfCES (liftHornClauses hornClauses), (trueTerm, falseTerm))
splitIf _ = Nothing
