module OrderedRewriting where

import Data.Maybe
import Data.List as DL

import ES
import Equation
import LPO
import Substitution
import TRS
import Term
import Util

type TermOrder = Term -> Term -> Bool

type OTRS = (ES, TermOrder)

-- extended critical pair
ecp' :: Equation -> Equation -> Position ->
        TermOrder -> Maybe (Term, Term)
ecp' e1 e2 p gt
-- gt is order > on terms
-- p is non-variable position of (l2, _) = e2
  | p == emptyPosition, variant e1 e2 = Nothing
  | Just theta <- unify l1 (subtermAt l2 p) =
      let [l1', r1', l2', r2'] = [ substitute t theta | t <- [l1, r1, l2, r2]]
      in if not (gt r1' l1') && not (gt r2' l2')
         then Just (replace l2' r1' p, r2')
         else Nothing
  | otherwise = Nothing
  where [(l1,r1), (l2,r2)] = rename 0 [e1, e2]

ecp :: OTRS -> [(Term, Term)]
ecp (es, gt) = catMaybes [ ecp' e1 e2 p gt | e1 <- es',
                                             e2@(l2, _) <- es',
                                             p <- functionPositions l2]
  where es' = es ++ (ES.inverse es) -- NOTE: don't use ES.union here

type VariableOrder = [String]

gtELPOlex :: Precedence -> VariableOrder -> Term -> [Term] -> [Term] -> Bool
-- gtELPOlex prec vgt s ss ts
-- assummption length ss = length ts
gtELPOlex _ _ _ [] [] = False
gtELPOlex _ _ _ ss [] = error "length ss = length ts does not hold."
gtELPOlex _ _ _ [] ts = error "length ss = length ts does not hold."
gtELPOlex prec vgt s (si : ss) (ti : ts)
  | si == ti = gtELPOlex prec vgt s ss ts
  | otherwise = gtELPO prec vgt si ti && all (\tj -> gtELPO prec vgt s tj) ts

-- extended lpo >_elpo
-- assume that prec is total precedence.
gtELPO :: Precedence -> VariableOrder -> Term -> Term -> Bool
gtELPO prec vgt (V x) (V y) = Util.gtByList vgt x y
gtELPO prec vgt s (V y) = elem y (variables s)
-- note that the 3 cases below are NOT distinct.
gtELPO prec vgt s@(F f ss) t@(F g ts)
  | LPO.gtPrec prec f g, all (\ti -> gtELPO prec vgt s ti) ts = True
  | any (\si -> si == t || gtELPO prec vgt si t) ss = True
  | f == g , gtELPOlex prec vgt s ss ts = True
  | otherwise = False
gtELPO _ _ _ _ = False

-- ordered rewriting

rewrite' :: Equation -> TermOrder -> Position -> Term -> Maybe Term
rewrite' (l, r) gt p s =
    case match l (subtermAt s p) of
      Just sigma -> let l' = substitute l sigma
                        r' = substitute r sigma
                    in if l' `gt` r' -- l > r
                       then Just (replace s r' p)
                       else Nothing
      Nothing -> Nothing

rewrite :: OTRS -> Term -> Maybe Term
rewrite (es, gt) s = case (catMaybes res) of x : _ -> Just x
                                             []    -> Nothing
  where res = [ rewrite' e gt p s
              -- union operation must be ++.
              | e <- (es  ++ (ES.inverse es)), p <- positions s]

nf :: OTRS -> Term -> Term
nf otrs t
  | Just t' <- rewrite otrs t = nf otrs t'
  | otherwise = t

joinable :: OTRS -> Term -> Term -> Bool
joinable otrs s t = (nf otrs s) == (nf otrs t)

-- ground joinability testing

-- `enumerateVSubs vs` enumerates variable subsitutions on vs.
enumerateVSubs :: [String] -> [Subst]
-- just lifts to substitutions.
enumerateVSubs vs = [ [(x, V y) | (x, y) <- f] | f <- Util.allFunctions vs vs]

-- returns True if s and t are joinable w.r.t. OTRS (E, >_lpo)
-- Note that False means MAYBE i.e. this function is sound but not complete.
groundJoinable :: Precedence -> ES -> Term -> Term -> Bool
groundJoinable prec es s t =
  all (\(rho, vord) -> joinable (es, gtELPO prec vord) (substitute s rho) (substitute t rho))
      [(rho, vord) | rho <- enumerateVSubs stvars, vord <- allTotalStrictOrders stvars]
  where stvars = DL.nub (variables s ++ variables t)

-- ground confluence testing

-- ground confluence testing for OTRS with LPO
-- Note that False means MAYBE i.e. this function is sound but not complete.
groundConfluent :: Precedence -> ES -> Bool
groundConfluent prec es = all (\(s, t) -> groundJoinable prec es s t) (ecp (es, gtLPO prec))

-- NOTE: we cannot use list comprehension.
-- delete and simplify must be performed sequentially.
interreduce' :: Precedence -> ES -> ES -> ES
interreduce' prec [] reduced = reduced
interreduce' prec ((l, r) : es) reduced
  | groundJoinable prec (ES.diff es' [(l', r')]) l' r' = interreduce' prec es reduced
  | otherwise = interreduce' prec es (ES.union [(l', r')] reduced)
  where es' = ES.diff (ES.union es reduced) [(l, r)]
        otrs = (es', gtLPO prec)
        l' = nf otrs l
        r' = nf otrs r

-- args: (E, >)
-- performs "simplify" and "delete"
interreduce :: Precedence -> ES -> ES
interreduce prec es = interreduce' prec es []
