module Prover (prover) where

import Data.List

import Equation
import ES
import LPO
import SMT
import Term
import LPOAuto
import OrderedRewriting
import Substitution
import Horn

findPrec :: ES -> IO Precedence
findPrec es = do
  res <- solve "z3" "Prover" z3Input
  case res of
    Nothing -> undefined
    Just res' -> let prec = map fst (sortOn snd [ (fname, -n) | (fname, v) <- res', n <- fromSMTValue2Int v])
                 in return prec
  where
    esWithIds = [("rulevar" ++ show i, e) | (i, e) <- zip [0..] (es ++ ES.inverse es)]
    functionSymbols = nub [f | (l, r) <- es, f <- (Term.functions l) ++ (Term.functions r)]
    esIds = [ esId | (esId, _) <- esWithIds ]
    genReducibilityFormula (s, t) =
      orFormula [ BoolVar ruleId | (ruleId, (l, _)) <- esWithIds,
                                   encompass s l || encompass t l]
    -- maximize the # of reducible equations
    maximizedConstraints = [ genReducibilityFormula e | e <- es ]
    lpoDic = constructDic (map snd esWithIds)
    dicCostraints = andFormula [ impliesFormula (BoolVar name) (encodeLPODic lpoDic s t) | ((s, t), name) <- lpoDic]
    orientation = andFormula [ impliesFormula (BoolVar ruleId) (BoolVar (unsafeLookup (l, r) lpoDic))
                             | (ruleId, (l, r)) <- esWithIds ]
    -- totality is required since precedence is presented as list
    precTotality = Distinct [Var f | f <- functionSymbols]
    z3Input = maxsat functionSymbols
                     (esIds ++ map snd lpoDic)
                     maximizedConstraints
                     (andFormula [orientation, precTotality, dicCostraints])

generateLemmata :: (ES, Precedence) -> ES
generateLemmata (es, prec) = selectLemmata (ES.diff cps es)
  where cps = [ e | e@(s, t) <- ecp (es, gtLPO prec),
                    not (groundJoinable prec es s t) ]

selectLemmata :: ES -> ES
selectLemmata es = take n (ES.sortByWeight es)
  where n = 12

trueEqualsFalse :: Equation
trueEqualsFalse = (trueTerm, falseTerm)

-- True means "unsat"
-- False means "sat"
prover' :: ES -> ES -> IO Bool
prover' initES es = do
  putStrLn "current es:"
  putStrLn (showES es)
  prec <- findPrec es
  putStrLn ("found prec: " ++ LPO.showPrec prec)
  let es' = OrderedRewriting.interreduce prec es
  putStrLn "interreduced:"
  putStrLn (showES es')
  let lemmata = generateLemmata (es', prec)
  putStrLn "new lemmata:"
  putStrLn (showES lemmata)
  let nextES = (ES.union es' lemmata)
  -- First, check if true = false is derived.
  -- This must be done before we check ground completeness.
  if any (\e -> Equation.variant e trueEqualsFalse) nextES
    then return True
    else if all (\(s, t) -> groundJoinable prec es' s t) -- ground completeness checking
                (ES.union (ecp (es', gtLPO prec)) initES)
          then return False
          else prover' initES nextES

-- True means "unsat"
-- False means "sat"
prover :: ES -> IO Bool
prover es = prover' es es
