module SMT where

import Data.List
import System.Process
import System.IO
import Text.ParserCombinators.Parsec
import Control.Monad

data Exp = Var String
         | Val Int
         | Plus [Exp]
         | Times [Exp]
         deriving Eq

data Formula = And [Formula]
             | Or [Formula]
             | Not Formula
             | Distinct [Exp]
             | Gt Exp Exp
             | Geq Exp Exp
             | Gtn [Exp] -- (> t1 t2 t3 ... tn)
             | Eq Exp Exp
             | BoolVar String
             deriving Eq

instance Show Exp where
    show (Var s)     = s 
    show (Val n)       = show n
    show (Plus []) = "0"
    show (Plus [e]) = show e
    show (Plus es)  = "(+ " ++ intercalate " " [show e | e <- es] ++ ")"
    show (Times []) = "1"
    show (Times [e]) = show e
    show (Times es)  = "(* " ++ intercalate " " [show e | e <- es] ++ ")"

instance Show Formula where
    show (BoolVar s) = s
    show (And []) = "true"
    show (And [f]) = show f
    show (And fs) = "(and " ++ intercalate " " [show f | f <- fs] ++ ")"
    show (Or []) = "false"
    show (Or [f]) = show f
    show (Or fs) = "(or " ++ intercalate " " [show f | f <- fs] ++ ")"
    show (Not f) = "(not " ++ show f ++ ")"
    show (Distinct es) =  "(distinct " ++ intercalate " " [show e | e <- es] ++ ")"
    show (Gt e1 e2) = "(> " ++ show e1 ++ " " ++ show e2 ++ ")"
    show (Geq e1 e2) = "(>= " ++ show e1 ++ " " ++ show e2 ++ ")"
    show (Gtn es) =  "(> " ++ intercalate " " [show e | e <- es] ++ ")"
    show (Eq e1 e2) = "(= " ++ show e1 ++ " " ++ show e2 ++ ")"

-- smart constructors

andFormula :: [Formula] -> Formula
andFormula [f] = f -- propagete partial evaluation
andFormula fs
  | falseFormula `elem` fs' = falseFormula
  | otherwise = case fs' of
                  [] -> trueFormula
                  _  -> And fs'
  where fs' = [f | f <- fs, f /= trueFormula]

orFormula :: [Formula] -> Formula
orFormula [f] = f -- propagate partial evaluation
orFormula fs
  | trueFormula `elem` fs' = trueFormula
  | otherwise = case fs' of
                  [] -> falseFormula
                  _  -> Or fs'
  where fs' = [f | f <- fs, f /= falseFormula]

notFormula :: Formula -> Formula
notFormula f
  | f == trueFormula = falseFormula
  | f == falseFormula = trueFormula
  | otherwise = Not f

impliesFormula :: Formula -> Formula -> Formula
impliesFormula p q = orFormula [notFormula p, q]

trueFormula :: Formula
trueFormula = And []

falseFormula :: Formula
falseFormula = Or []

data Command =
     DeclareInt String
   | DeclareBool String
   | Assert Formula
   | AssertSoft Formula
   | CheckSat
   | GetValue [String]

type SMTInput = [Command]

data SMTValue = SMTBoolValue Bool | SMTIntValue Int
                deriving Show

isSMTBoolTrue :: SMTValue -> Bool
isSMTBoolTrue (SMTBoolValue True) = True
isSMTBoolTrue _ = False

fromSMTValue2Int :: SMTValue -> [Int]
fromSMTValue2Int (SMTIntValue n) = [n]
fromSMTValue2Int _ = []

smtValueUnsafeCastInt :: SMTValue -> Int
smtValueUnsafeCastInt (SMTIntValue i) = i
smtValueUnsafeCastInt (SMTBoolValue True) = 1
smtValueUnsafeCastInt (SMTBoolValue False) = 0

type SMTOutput = Maybe [(String, SMTValue)]

instance Show Command where
    show (DeclareInt varname) = "(declare-const "  ++ varname ++ " Int)"
    show (DeclareBool varname) = "(declare-const "  ++ varname ++ " Bool)"
    show (Assert f) = "(assert " ++ show f ++ ")"
    show (AssertSoft f) = "(assert-soft " ++ show f ++ ")"
    show CheckSat = "(check-sat)"
    show (GetValue vars) = "(get-value (" ++ intercalate " " [ v | v <- vars ] ++ "))"

showSMTInput :: SMTInput -> String
showSMTInput commands = intercalate "\n" [show c | c <- commands]

generateSMTInput :: [String] -> [String] -> Formula -> SMTInput
generateSMTInput ivs bvs formula = commands
  where commands =  [DeclareInt v | v <- ivs]
                 ++ [DeclareBool v | v <- bvs]
                 ++ [Assert formula, CheckSat, GetValue ivs ]

maxsat :: [String] -> [String] -> [Formula] -> Formula -> SMTInput
maxsat ivs bvs ss h = commands
  where commands =  [DeclareInt v | v <- ivs]
                 ++ [DeclareBool v | v <- bvs]
                 ++ [AssertSoft s | s <- ss]
                 ++ [Assert h, CheckSat, GetValue (ivs ++ bvs)]

parseSMTOutput :: Parser SMTOutput
parseSMTOutput = try parseSat <|> parseUnsat

parseSat :: Parser SMTOutput
parseSat = do
  spaces
  void $ string "sat"
  spaces
  void $ char '('
  spaces
  bindings <- many parseBinding
  spaces
  void $ char ')'
  spaces
  return $ Just bindings

parsePosInt :: Parser SMTValue
parsePosInt = do
  spaces
  i <- many1 digit
  spaces
  return $ SMTIntValue $ read i

parseNegInt :: Parser SMTValue
parseNegInt = do
  spaces
  void $ char '('
  spaces
  void $ char '-'
  spaces
  i <- many1 digit
  spaces
  void $ char ')'
  spaces
  return $ SMTIntValue $ read i

parseInt :: Parser SMTValue
parseInt = try parsePosInt <|> parseNegInt

parseTrue :: Parser SMTValue
parseTrue = do
  void $ string "true"
  return $ SMTBoolValue True

parseFalse :: Parser SMTValue
parseFalse = do
  void $ string "false"
  return $ SMTBoolValue False

parseBool :: Parser SMTValue
parseBool = parseTrue <|> parseFalse

parseBinding :: Parser (String, SMTValue)
parseBinding = do
  spaces
  void $ char '('
  spaces
  varname <- manyTill (noneOf " \t\r\n") space
  spaces
  val <- parseInt <|> parseBool
  spaces
  void $ char ')'
  spaces
  return (varname, val)

parseUnsat :: Parser SMTOutput
parseUnsat = do
  spaces
  void $ string "unsat"
  spaces
  return Nothing

solve :: String -> String -> SMTInput -> IO SMTOutput
solve toolPath tmpFileName input = do
  let z3InputFileName = "/tmp/" ++ tmpFileName ++ ".smt2" 
  writeFile z3InputFileName $ showSMTInput input
  (_, Just hout, _, _) <-
      createProcess (proc toolPath [z3InputFileName]){ std_out = CreatePipe }
  z3Output <- hGetContents hout
  case parse parseSMTOutput "z3Output" z3Output of
      Right r -> return r
      Left e -> error $ show e
